/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
     author: kekec 		E-mail: behaviour.brain.research@gmail.com
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

 
//                          lpt version: 1.1


//  xxxxxxxxxxxxxxx    global variable declarations  xxxxxxxxxxxxxxxx

 
// These variables below can be directly modified by the user on the options form: 

var c_min = 10;  // Minimum number of generated characters (columns)

var c_max = 10; // Maximum number of generated characters (columns)

var iter = 3;   // Number of iterations for each row value

var is_lower = '3'; // Letter type: '1' = Original char set from text file, '2' = lowercase; '3' = UPPERCASE

var is_req = false; // Some letters are always displayed: '1' (No) => false. '2' (Yes) => true

var is_ltr = '1';   // Parsing (reading) direction: '1' = left-to-right, '2' = right-to-left, '3' = both. see getScore()

var limen = 1;  // Minimum word length to be accepted as valid (threshold)

var duplic = false; // It is allowed to repeat the same letters within the same item (true || false)

var is_sort = true; // Sort ascending letters (true || false)

var timer_enabled = false;   // if true => timer is enabled. if false timer is disabled. Exactly: the timer is always on

var timer = 600;     // Task time in seconds if clock = 1000ms (default). 

var font_family = 'Arial, Helvetica, sans-serif';  // default canvas ctx.fontFamily 

var is_back = true;    // if (is_back==true) => user can go back to previous task/item. (true || false)

var prg_type = '0';   //  type of information in Progress Bar ('0' => Items, '1' => Time; '2' => Nothing) 

var is_shuffle = false;   // Randomly shuffle rows (only when Demo mode = Off)

var demo = true;       // if (demo == true) => Show task solutions 

var bclr_d = '#fff';  // Top line background color

var pclr_d = '#000';  // Font color - for top line

var bclr_s = '#e5e5e5'; 		// Bottom line background color				
 
var pclr_s = '#000';   // Font color - for bottom line

var is_border = true;   // Draw border around numbers

var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 

// Other global variables:

const clock = 1000;  // basic timer resolution in ms

var timer_id = null;   // timer_id is the identifier of the repeated action you want to cancel. This ID is returned from setInterval().

var counter = 0;      // timer counter. if (counter >= time) Timer will call theEnd

var item_cnt;       // number of items in the task. range [0..item_cnt-1]. see setPar()

var item;   				// current item in task

var is_task = false;    // task in progress or not for pageResize() event. It is true for all events between end of startTask() and theEnd() events

var dirty = false;  	// if  (dirty==true) => user changed any options in form 

var is_storage = false;		// local storage is enabled on user browser or not
  
//  global objects and arrays declarations..

var letters = { a:1,  b:1,  c:1,  d:1,  e:1,  f:1,  g:1,  h:1,  i:1,  j:1,  k:1,  l:1,  m:1,  n:1,  o:1,  p:1,  q:1,  r:1,  s:1,  t:1,  u:1,  v:1,  w:1,  x:1,  y:1,  z:1 };  // Alphabetic letters object. Default: English alphabet (Latin alphabet consisting of 26 letters)

var char_def = true;  // default Alphabetic letters array is used (true) or not (user loaded custom letters from txt file) 

var dic_def = true;  // default Dictionary is used (true) or not (user loaded custom Dictionary loaded from txt file) 

var stim = [];   // stimulus array. Dimensions: [items] x [character count in this item]. see stimGen()   

var ans = [];    // user answer array. Dimensions: [items] x [character count in this item].

var rndKey = function (obj) {
    let keys = Object.keys(obj);
    return keys[ keys.length * Math.random() << 0];
};

/*
var rndValue = function (obj) {
    let keys = Object.keys(obj);
    return obj[keys[ keys.length * Math.random() << 0]];
}; */



/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */



// application init [create/start Timer, check storage, show start panel..]
function init() 
{ 
  if (timer_id)  {   
    window.clearInterval(timer_id);  
  }  
  timer_id= setInterval(function () {appTimer()},clock);  
	
	is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }	
		
	loadFont();
	
	setIns();
	
	document.getElementById('letters_cnt').textContent= Object.keys(letters).length;  // Number of letters of the alphabet (26 or loaded from txt file) 
	
	document.getElementById('words_cnt').textContent= Object.keys(words).length;  // Number of words loaded from the dictionary
	
	setChar();
	
	showStart();  
}

 
// application timer handler function
function appTimer()  
{    
	if (timer_enabled && is_task)   // else: nop, idle, do nothing 
	{
		counter++;  
		if (counter >= timer)  {     // task time is limited
			theEnd(); 
		}	
		else if (prg_type == '1')   {   // '1' => Time. display time on Progress Bar  
			let rest= timer-counter;   // this is remaining time 
			let percent= Math.round(100*rest/timer);    
			document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
			document.getElementById('nav_items').textContent= stringTime(rest);	
		}
	}  
}	


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 (probably redundant)
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// return integer into string hh:mm:ss time format
// input par: total == integer (time in seconds)
function stringTime(total) 
{
  let hh = Math.floor(total/(60*60));         
  let div_min = total % (60 * 60);
  let mm = Math.floor(div_min / 60);       
  let div_sec = div_min % 60;
  let ss = Math.ceil(div_sec);    
  // These lines ensure you always have two-digits. This formats your string to HH:MM:SS
  if (hh < 10)  {hh = "0"+hh;}     
  if (mm < 10) {mm = "0"+mm;}
  if (ss < 10) {ss = "0"+ss;} 
  return hh + ":" + mm + ":" + ss;
} 


// fill font-family names from array into select html element
function loadFont()  
{
	// used font-family array: The name of a font family. For example, "Times" and "Helvetica" are font families. Font family names containing whitespace should be quoted
	// https://www.w3schools.com/cssref/css_websafe_fonts.asp	
	const arr = ['"Times New Roman", Times, serif', 'Georgia, serif', '"Palatino Linotype", "Book Antiqua", Palatino, serif', 'Arial, Helvetica, sans-serif', '"Arial Black", Gadget, sans-serif', '"Comic Sans MS", cursive, sans-serif', 'Impact, Charcoal, sans-serif', '"Lucida Sans Unicode", "Lucida Grande", sans-serif', 'Tahoma, Geneva, sans-serif', '"Trebuchet MS", Helvetica, sans-serif', 'Verdana, Geneva, sans-serif', '"Courier New", Courier, monospace', '"Lucida Console", Monaco, monospace' ];
	let sel= document.getElementById('font_family'), len = arr.length, opt; 
	sel.innerHtML= ''; 
	for (let i= 0; i!= len; i++) 
	{  
		opt = document.createElement('option');
		opt.value= arr[i];  // option value selected font-family name (string)            
		opt.text= arr[i];   // option textContent => font-family name (string)              		                       								
		sel.add(opt);  // sel.add(opt, null);                 			
	} 
	sel.selectedIndex= 3; // 'Arial, Helvetica, sans-serif'
	document.getElementById('font_family_msg').style.fontFamily= sel.value; // = arr[3];
}	


// raised when page is resized. see html: <body onresize="pageResize()"
function pageResize()  
{ 
	if ( is_task )  { 
		document.getElementById('can_m').style.display= 'none';  // tmp until mouseMove()
		setCanvas();		
		drawStim();		
	}
}	


// show start panel 
function showStart()  
{     
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('hide_option').style= 'text-align:center';  // bug? auto-remove style when element.style.display="none"
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (is_storage && localStorage.getItem('lpt-v1.1')) 
  {
    let cnt= JSON.parse(localStorage.getItem('lpt-v1.1')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';  // xxx
		}
	}
  dirty= false;
  //document.getElementById('is_valid').textContent='';  // tmp reset validator message
	document.getElementById('no_task').style.display= 'none';  // specific to this app
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').classList.add('center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';          
  document.getElementById('hide_option').style.display = 'none';   
  //document.getElementById('start_div').style= 'text-align:center';  // bug?  (firefox auto remove class "center" when set style.display.none;)
  document.getElementById('show_option').style= 'text-align:center';  
  document.getElementById('instructions').style.display= 'block';
	refMsg(); 
  setIns();	
}


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		

	document.getElementById('ins_back').style.display= is_back ? 'none' : 'block';
	
	if ( document.getElementById('timer_enabled').checked )
  {   
    document.getElementById('ins_timer').setAttribute('style', 'display:block'); 		
		document.getElementById('ins_timer').textContent= 'The duration of this task is limited.';  // The duration of this task is always limited.
		if (prg_type == '0')  {
			document.getElementById('ins_timer').textContent+= ' The Progress bar will show how many items are left until the end of the task.';
		}
		else if (prg_type == '1')  {
			document.getElementById('ins_timer').textContent+= ' The Progress bar will display the remaining time until the end of the task.';  
		}
		else  {
			document.getElementById('ins_timer').textContent+= ' But information about the remaining time will not be visible.';  
		}  
  }
	else  {
		document.getElementById('ins_timer').setAttribute('style', 'display:none');
	}	

	if ( limen > 1 ) {
		document.getElementById('ins_limen').style.display= 'block';
    document.getElementById('ins_limen').textContent= 'If you want your word to be valid, the minimum number of letters used must be: ' + limen;
	}
	else  {
		document.getElementById('ins_limen').style.display= 'none';	
	}
	document.getElementById('ins_demo').style.display= demo ? 'block' : 'none';	
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty = true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{
		
		case 'c_min':
		case 'c_max':
			if (e.validity.valid)  {
				msg.textContent= ''; 				
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
			} 
			checkItem();
			break; 
   	  
		case 'iter':
			if (e.validity.valid)  {
				msg.textContent= 'Number of repetitions for each length'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			checkItem();
			break; 	
			    
		case 'is_lower':	
			switch ( e.value )	{
				case '1':
					msg.textContent= 'The text renders as it is. When using your own stimuli, probably the most reliable option is to choose the original.';
					break;
				case '2':
					msg.textContent= 'Lowercase characters will always be displayed on the screen.';
					break;	
				case '3':
					msg.textContent= 'UPPERCASE characters will always be displayed on the screen.';
					break;		
      }     
			break; 
		
		case 'is_req':	 
		  if (e.value == '1') {					
				msg.textContent= 'The choice of letters is random.';
				document.getElementById('d_is_req').style.display= 'none';
			} 
      else  {
				msg.textContent= 'The letters you select will always be displayed.';
				document.getElementById('d_is_req').style.display= 'block';
			}				
			break;
		
		case 'is_ltr':		
		  switch ( e.value ) {
				case '1':
			    msg.textContent= 'Default. Left-to-right text direction.';
					break;
			  case '2':
			    msg.textContent= 'Right-to-left text direction.';
					break;
				case '3':
			    msg.textContent= 'The original and reverse answers are equivalent. Eg if the user chooses the word "example" OR "elpmaxe" both are valid.';
				break;			       
			}
			msg.textContent += ' Applies only to parsing text in the top line.';
			break; 
		
		case 'limen':
			if (e.validity.valid)  {
				msg.textContent= 'This is the minimum validity threshold. If you set this value greater than 1, all participant answers that are shorter than the selected one will always be invalid (0 points), regardless of whether the word is in the dictionary or not.'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break; 	
		
		/*case 'duplic':		
			msg.textContent= e.value == '1' ? '' : '';      
			break; */
		
		case 'is_sort': 
			msg.textContent= e.checked ? 'Yes. We recommend using sorting only if you use the default Alphabet, and left-to-right text direction.' : 'No.';  
			break; 
				
		case 'timer_enabled':   
      if (e.checked)  {
        msg.textContent= 'Yes.';
        document.getElementById('p_timer').style.display= 'block';
      }
      else  {  
        msg.textContent= 'No. Timer is disabled.';
        document.getElementById('p_timer').style.display= 'none'; 
      }       
      break;
		
    case 'timer':
			if (e.validity.valid)  {
				msg.textContent= stringTime(e.value); 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break; 

		/* case 'font_family':
		  break; */
		
		case 'is_back':		
			msg.textContent= e.checked ? 'Yes.' : 'No.';      
			break; 
						
		case 'prg_type':                                                                                                                                            
      if (e.value == '0')  {
        msg.textContent= 'It will be displayed how many items are left to the end.'; // The number of the current item will be displayed, as well as the total number of items.
      }
      else if (e.value == '1')  {
        msg.textContent= 'The remaining time until the end of the task will be displayed. Only when timer is enabled.';
      }
      else  {
        msg.textContent= ''; //'Nothing';
      }
      break;    

    case 'is_shuffle':		
			msg.textContent= e.checked ? 'Yes. All stimuli will be randomly mixed. The ordinal number of the item and the length of the letters within the item will not be directly correlated.' : 'No. Stimuli with a smaller number of letters will be displayed first. Then the number of letters will increase up to the maximum length.';      
			break; 

		case 'demo': 
			msg.textContent= e.checked ? 'On. If you click the Check button at the bottom of the screen, you will receive feedback on the validity of your word.' : 'Off';  
			break; 
	
		case 'bclr_d': 
		case 'pclr_d': 
    case 'bclr_s': 
		case 'pclr_s': 
			msg.textContent= validColor(e.value)  ? '' : 'The selected color is invalid.'; 
			break; 	

    case 'is_border':		
			msg.textContent= e.checked ? 'Yes.' : 'No.';      
			break; 
			
		case 'c_del_l':		// Load the letters of the alphabet. Delimiter character
			document.getElementById('s_del_l').style.display= e.checked ? 'none' : 'inline';     
			break; 
		
    case 'c_del_w':	// 	Load the dictionary. Delimiter character
			document.getElementById('s_del_w').style.display= e.checked ? 'none' : 'inline';     
			break;

		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// check if item count is valid or not
// update item count on options form and check c_min <= c_max and update item_total label
function checkItem()  
{
  let c_min = document.getElementById('c_min');
  let c_max = document.getElementById('c_max');
  let iter = document.getElementById('iter');
  let items = document.getElementById('item_total');
  let c_min_msg = document.getElementById('c_min_msg');
  let c_max_msg = document.getElementById('c_max_msg');    
  if (c_min.validity.valid && c_max.validity.valid && iter.validity.valid)   
  {                      
    if  ( parseInt(c_min.value,10) > parseInt(c_max.value,10) )  
    {  
			let tmp= 'Minimum number of letters must not be higher than the Maximum number of letters.';
      c_min_msg.textContent= tmp;  
      c_max_msg.textContent= tmp;  
      items.textContent= 'Total number of items = Invalid value'; 
      items.setAttribute('class', 'error');
    } 
    else   
    {                                                                                               
      c_min_msg.textContent= '';  
      c_max_msg.textContent= '';
      let cnt= 0, level= parseInt(c_min.value,10);  // update cnt => total item count in task  
      while (level <= parseInt(c_max.value,10))  {
        cnt+= parseInt(iter.value,10);    
        level++;   
      } 
      items.textContent=  'Total number of items = ' + cnt ;      
      items.removeAttribute('class');
    }
  } 
  else  
  {
		let s = 'Values within this range are allowed: ', s2= ' with step= ';    
    c_min_msg.textContent= s + c_min.min + ".." + c_min.max + s2 + c_min.step;
		c_max_msg.textContent= s + c_max.min + ".." + c_max.max + s2 + c_max.step;
		items.textContent= 'Total number of items = Invalid value'; 
    items.setAttribute('class', 'error');
  }    
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty === true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity most of options 
// for other options see setPar()
function optionsValid()  
{ 
  let res = document.getElementById('c_min').validity.valid 
				 && document.getElementById('c_max').validity.valid
	       && document.getElementById('iter').validity.valid 			 
				 && document.getElementById('limen').validity.valid 
				 && document.getElementById('timer').validity.valid 
				 && validColor(document.getElementById('bclr_d').value)
				 && validColor(document.getElementById('pclr_d').value)
				 && validColor(document.getElementById('bclr_s').value)
				 && validColor(document.getElementById('pclr_s').value); 
	
	if ( parseInt(document.getElementById('c_min').value,10) > parseInt(document.getElementById('c_max').value, 10) )  {  
		res = false;
	}		
  return res;              
}


// start task (research) 
function startTask(e)   
{      			 	
  if ( setPar() == true )  {		
		
		initLayout(); 

		stimGen();
		
		is_task= true;   // if ( timer_enabled ) finally "activate" timer 
		
		drawStim();    // draw first item (stimulus)
	}
	// else  do nothing. wait for user to change settings
}


// init DOM/html elements (positions,visibility,..)
function initLayout()
{	
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';              
	document.getElementById('task_section').style.display= 'block'; 	  	
	
 	document.getElementById('progress').style.display= 'block';   //document.getElementById('progress').setAttribute('style', 'width:100%');	 
	if ( document.getElementById('prg_type').value == '2' || ( document.getElementById('prg_type').value == '1' && timer_enabled == false ) ) {
		document.getElementById('progress_holder').style.display= 'none'; 
	}	
	else  
	{
		document.getElementById('progress_holder').style.display= 'block';     
		//document.getElementById('progress').setAttribute('style', 'width:100%');   		
		if (document.getElementById('prg_type').value == '0') {
			document.getElementById('progress').className='progress-bar progress-bar-info progress-bar-striped';     //  items / remaining items
		}
		else if (document.getElementById('prg_type').value == '1')  {
			document.getElementById('progress').className= 'progress-bar progress-bar-danger progress-bar-striped';  // remaining time   
		} 	
		document.getElementById('progress').setAttribute('style', 'width:0');  // reset progress bar width
	}					
	document.getElementById('nav_items').textContent='';  // always reset  
  //document.getElementById('feedback').textContent='';  // in drawStim()
	
	document.getElementById('demo_div').style.display= demo ? 'block' : 'none';   // Check button visible or not
	
  setCanvas();
	
}  


// set html canvas element absolute positions
function setCanvas()
{

	// canvas must be after set footer content => progress_holder visible or not	
	let y = innerHeight - document.getElementById('footer').offsetHeight;
	document.getElementById('footer').top= y;
	
	// 1. set destination canvas (top/upper part of screen)
	let c= document.getElementById('can_d'), ctx= c.getContext('2d');
	c.style.display = 'block';   

	c.top= 0; 
	c.style.top = '0px';	
	
	c.width = innerWidth;  
	c.style.width = c.width + 'px';	     
	
	c.height = y/2;   
	c.style.height = c.height + 'px';  
	
	ctx.textAlign = 'center';    // ctx.textAlign = "left" || "right" || "center" || "start" || "end";   Default value: 	start
	ctx.textBaseline = 'middle';  // const baselines = ['top', 'hanging', 'middle', 'alphabetic', 'ideographic', 'bottom'];  Default value: alphabetic
  
	ctx.lineWidth= 1; 
	//ctx.strokeStyle= pclr_d;  // for draw border in drawStim() => equal to fillStyle.
	
	// 2. set source canvas (bottom/down part of screen)
	c= document.getElementById('can_s'), ctx= c.getContext('2d');
	c.style.display = 'block';   
	
	c.top= y/2; 
	c.style.top = c.top + 'px';	
	
	c.width = innerWidth;  
	c.style.width = c.width + 'px';	     
	
	c.height = y/2;  
	c.style.height = c.height + 'px';  
	
	ctx.textAlign = 'center';    // ctx.textAlign = "left" || "right" || "center" || "start" || "end";   Default value: 	start
	ctx.textBaseline = 'middle';  // const baselines = ['top', 'hanging', 'middle', 'alphabetic', 'ideographic', 'bottom'];  Default value: alphabetic
 	
  ctx.lineWidth= 1;
 	//ctx.strokeStyle= pclr_s;  // for draw border in drawStim() => equal to fillStyle.
	
}


/* set internal application parameters (global vars)
	return true if task can be started (or false if not)
  set content of id="no_task" validator message  */
function setPar()  
{  
	let ok = true;  // function result
	
	if ( c_min > c_max ) {  // this should be impossible but because of danger of indefinite loop in stimGen()
    let v = Math.min(c_min, c_max);
		c_min= v;
		c_max= v;
		document.getElementById('c_min').value= v;
		document.getElementById('c_max').value= v;
	}	

  document.getElementById('no_task').innerHTML= '';

  if ( limen > Math.min(c_min, c_max) ) {  
		document.getElementById('no_task').innerHTML = 'Minimum word length to be accepted as valid (threshold) is too big. The maximum allowed value is = ' + Math.min(c_min, c_max) + '. Currently set to = ' + limen + '.<br>';
    ok= false;
	}		
	
  if ( c_min < 3 || c_max < 3  || iter < 1 ) {
		document.getElementById('no_task').innerHTML+= 'Minimum (and Maximum) number of letters used must be greater than 2. The number of iterations must be greater than 0.<br>';
		ok= false;
	}

  // this is probably redundant check (this section should never happen)
	let let_cnt = Object.keys(letters).length;  // Number of letters of the alphabet (26 or loaded from txt file) 
	if ( let_cnt < 1 ) {
		document.getElementById('no_task').innerHTML+= 'The number of alphabet letters loaded is zero. Check the uploaded file or reload the current document, ie press the reload button in your browser to reload the default settings.<br>';  	
		ok = false;
	}
  else if ( let_cnt > 247 ) { 
		document.getElementById('no_task').innerHTML+= 'The number of letters of the alphabet used must be less than 248 (Tamil=247).<br>';
		ok = false;		
	} 
	if ( c_min > let_cnt || c_max > let_cnt) {
		document.getElementById('no_task').innerHTML+= 'The maximum (and minimum) number of letters used must be less than the number of letters of the alphabet.<br>';
		ok = false;
	}	
	
	let wor_cnt= Object.keys(words).length;   // Number of words loaded from the dictionary	
	if ( wor_cnt == 0 ) {
		document.getElementById('no_task').innerHTML+= 'The number of words loaded from the dictionary is zero. Check the uploaded file or reload the current document, ie press the reload button in your browser to reload the default settings.<br>';
		ok = false;
	}
	if (	(c_max - c_min + 1) * iter > wor_cnt ) {
		document.getElementById('no_task').innerHTML+= 'The number of items is greater than the total number of words loaded from the dictionary.<br>';
		ok= false;		
	}
	
	if ( ok ) {  // init task() => startTask()
		item = 0; 
		counter = 0;      // reset timer counter		
		stim = [];	      // reset stimulus array 
		ans = [];         // reset user answer array 
		document.getElementById('no_task').style.display= 'none';
		return true;
	}
	else  {
		document.getElementById('no_task').innerHTML += 'Unable to start task.'
		document.getElementById('no_task').style.display= 'block';
		return false;
  }	
}


// generate all stimuli for each item in range 0..item_cnt-1	
function stimGen()	
{		
  let n_letters= {};  // window.letters transformed toLocaleLowerCase or toLocaleUpperCase or original 
	if ( is_lower == '2' )  {
		for (let key in letters) {
			n_letters[key.toLocaleLowerCase()] = 1;			
		}
	}
	else if ( is_lower == '3' )  {
		for (let key in letters) {  
			n_letters[key.toLocaleUpperCase()]= 1;
		}
	}
	else {
		n_letters= letters;
	}
	
	// stim = [items] x [character count in this item]
	// col => number of displayed characters on screen (also column count)
	let col= Math.min(c_min, c_max), end= Math.max(c_min, c_max);  
	while ( col <= end )  
	{		
		for (let i= 0; i< iter; i++) 
		{   
	    let txt, row= []; 
		  txt= genString(col, n_letters);
      let size= txt.length;			
			for (let j= 0; j < size; j++)  {
				row.push(txt[j]);				
			}
			stim.push(row);			
		}		
		col++;	
	} 	
	
	item_cnt= stim.length;   // set item_cnt => total items count in task 

	if (is_shuffle) {
		stim = shuffle(stim);
	}
  
	// after [shuffle] set empty user answer for each item. ans array have equal dimensions like stim array
	
	for (let i = 0; i < item_cnt; i++) 
	{
		let len= stim[i].length, row= [];
    for (let j = 0; j < len; j++)  {
			row.push('');  
		}	
		ans.push(row);   // ans = [items] x [character count in this item]
	}
}


/* generates a string from letters array
  cnt = number of characters to generate
  letters = window.letters object transformed toLocaleLowerCase() or toLocaleUpperCase() or original. depends of is_lower value
	return string (string.length = cnt) */
function genString(cnt, letters)
{ 
  let arr = [], res = "", is_duplic, key= {}, c;
  if ( cnt > Object.keys(letters).length ) {
		is_duplic= true;		 // indefinite loop danger  
	}
	else {
		is_duplic= duplic;
	}
		
	if ( is_req )  
	{	// first collect all required char
		let col = document.getElementById('char_req').getElementsByClassName('char-req'), len = col.length;  // live HTMLCollection: class name for required chars in parent containter 
			//let col = document.getElementById('char_req').querySelectorAll('input[type=checkbox]').length;			
    for (let j= 0; j < len; j++) 
		{
			let e = col[j];  // checkbox element (required chars)
      if ( e.checked === true && arr.length < cnt ) 
			{
				c = e.value;  // c=  e.getAttribute('data-value');
        if ( is_lower == '2' ) {					
					c = c.toLocaleLowerCase();       			
				}
        else if ( is_lower == '3' )  {
          c= c.toLocaleUpperCase();
				}					// else not using:  .toLocaleLowerCase() .toLocaleUpperCase();				
				
				arr.push(c); //arr.push(e.getAttribute('data-value'));
				key[c]= 1;
			}
		}			
		cnt-= arr.length;
	}

	if ( is_duplic ) 
	{ // duplicates are allowed
		for (let i = 0; i < cnt; i++)  {			   
			c = rndKey(letters);  
			arr.push(c);
		}  
	}
	else 		// each generated character must be different. Only for default letters
	{  
		for (let i = 0; i < cnt; i++)  
		{	
			do  {				    
				c = rndKey(letters); 
			}	while ( key[c] == 1 ); 
			key[c]= 1;
			arr.push(c);   		
		} 
	}
	
	if (is_sort) {
		arr.sort();
	}
	else  {
		arr = shuffle(arr);
	}
	
	res= arr.join('');  
	return res;
}

 
// unbiased shuffle algorithm: Fisher-Yates (aka Knuth) Shuffle.
function shuffle(arr) 
{
  let cur_index = arr.length, tmp, rnd_index;
  // While there remain elements to shuffle...
  while (0 !== cur_index) {    
    rnd_index = Math.floor(Math.random() * cur_index);     // Pick a remaining element...
    cur_index = cur_index - 1;   
    tmp = arr[cur_index];          // And swap it with the current element.
    arr[cur_index] = arr[rnd_index];
    arr[rnd_index] = tmp;
  }
  return arr;
}


// draw (repaint) current stimulus           
function drawStim() 
{    		 
  let can = document.getElementById('can_d'), ctx = can.getContext('2d');
	let col_cnt= stim[item].length;
	let width  = can.width / col_cnt;  // cell width => equal for id= 'can_s' and id= 'can_d'
	let height = can.height; // cell height => equal for id= 'can_s' and id= 'can_d'
	let r = Math.floor( Math.min(width,height)/2 );  // const
	ctx.font = r.toString() + 'px ' + font_family;   	// update font
	ctx.fillStyle= bclr_d;  // background-color top row
  ctx.fillRect(0, 0, can.width, height);
  ctx.fillStyle= pclr_d;  // 'font-color' - top row
	ctx.strokeStyle= pclr_d;  // for draw border only
	//let top= 0;
	for (let i = 0; i <= 1; i++)
	{		// i = 0 => draw top row. i = 1 => draw bottom row
			if ( i == 1 ) {  // draw bottom row. change canvas object reference
				can= document.getElementById('can_s');
				ctx = can.getContext('2d');	
        ctx.font = r.toString() + 'px ' + font_family;   	// must update font for different ctx	
        ctx.fillStyle= bclr_s;  // backgorund-color bottom row
				ctx.fillRect(0, 0, can.width, can.height);
 			  ctx.fillStyle= pclr_s;  // 'font-color' - bottom row
				if ( is_border ) {
					ctx.strokeStyle= pclr_s; 
				}
			}			
			for (let col= 0; col < col_cnt; col++)  
			{                						
				let left= col * width;   // cell left bound     
				let txt = i == 0 ? ans[item][col] : stim[item][col];								
				ctx.fillText(txt, Math.round(left+width/2), Math.round(height/2));	 //ctx.fillText(txt, Math.round(left+width/2), Math.round(top+height/2));		
				if ( is_border )	{
					ctx.strokeRect(left, 0, width, height);
					//ctx.beginPath();  
					//ctx.rect(left, 0, width, height);     
					//ctx.stroke(); 	
				}			
			}
	} 

	updateFooter();	
	document.getElementById('feedback').textContent= '';  // always reset
}


// raised when user change font-family on html <select> element
function tryFont(e)  {	
	document.getElementById('font_family_msg').style.fontFamily= e.value;
}	


// Changes application settings that can be changed directly by the user within the options form
// These variables below can be directly modified by the user on the options form
function setVarPar()  
{   
	
	if (document.getElementById('c_min').validity.valid && document.getElementById('c_max').validity.valid 
    && (parseInt(document.getElementById('c_min').value,10) <= parseInt(document.getElementById('c_max').value, 10) ) )   {  
      c_min= parseInt(document.getElementById('c_min').value, 10); 
      c_max= parseInt(document.getElementById('c_max').value, 10); 
  }	

	if (document.getElementById('iter').validity.valid)  {  
	  iter = parseInt(document.getElementById('iter').value, 10);			
  } 	
	
	is_lower= document.getElementById('is_lower').value; 
	
	is_req= document.getElementById('is_req').value == '2' ? true : false;  // Some letters are always displayed:'2' (Yes) => true, '1' (No) => false. 
		
	is_ltr= document.getElementById('is_ltr').value;
	
	if ( document.getElementById('limen').validity.valid ) {
		limen= parseInt(document.getElementById('limen').value, 10);
	}
	
	duplic= document.getElementById('duplic').value == '1' ? false : true;  // '1' = No, '2' = Yes
	
	is_sort= document.getElementById('is_sort').checked ? true : false;		
	
  timer_enabled= document.getElementById('timer_enabled').checked ? true : false;  // timer is enabled or disabled	
	
	if (timer_enabled && document.getElementById('timer').validity.valid) {
		timer= Math.round(parseInt(document.getElementById('timer').value, 10));  // task time (s)
  }	
		
	font_family= document.getElementById('font_family').value;	
  
	is_back= document.getElementById('is_back').checked ? true : false; 
	
	prg_type= document.getElementById('prg_type').value;  
	
  is_shuffle= document.getElementById('is_shuffle').checked ? true : false;   
	
  demo= document.getElementById('demo').checked ? true : false;	
	
	if (validColor(document.getElementById('bclr_d').value))  {		
		bclr_d= document.getElementById('bclr_d').value;
  }
	
	if (validColor(document.getElementById('pclr_d').value))  {		
		pclr_d= document.getElementById('pclr_d').value;
  }
	
	if (validColor(document.getElementById('bclr_s').value))  {		
		bclr_s= document.getElementById('bclr_s').value;
  }
	
	if (validColor(document.getElementById('pclr_s').value))  {		
		pclr_s= document.getElementById('pclr_s').value;
  }
		
  is_border = document.getElementById('is_border').checked ? true : false;
	
  store= (is_storage && document.getElementById('store').value === '1') ? true : false;
  
}


// test color for validity (from string)
// https://stackoverflow.com/questions/8027423/how-to-check-if-a-string-is-a-valid-hex-color-representation/8027444
function validColor(clr){
  return /^#[0-9A-F]{6}$/i.test(clr);
}


// goto previous item (stimulus)
function priorItem()  
{      
  if (item > 0)  {
    document.getElementById('can_m').style.display= 'none';  // raster can be changed in next/prior task
		item--;  
    drawStim();     // draw again previous stimulus
  }         
}


// goto next item (stimulus)
function nextItem()  
{    		
	item++;					
	document.getElementById('can_m').style.display= 'none'; // raster can be changed in next/prior task
	if (item < item_cnt)  {         		 
    drawStim();     // draw next stimulus => range 0..item_cnt-1        
  }
  else  {               
    theEnd();  // end of task if item >= item_cnt 
  }
}


// update footer when show each stimulus in task. update buttons and update progress bar
function updateFooter()
{  
	document.getElementById('prior').style.display= item > 0 && is_back ? 'inline': 'none';  // set footer navigation
	let btn= document.getElementById('next');     
	if (item == item_cnt-1)  {       
		btn.style.backgroundImage=  "url('img/end.png')";  // update 'Next item' image on 'right arrow' button 
		btn.title= 'End of task';
	}
	else  {
		btn.style.backgroundImage= "url('img/forward.png')"; 
		btn.title= 'Go forward';
	}						
  if (prg_type == '0' )  {    // update Progress bar => show current item count and remaining item count on Progress Bar                                                                                                                                                       
		let percent = Math.round(100*(item_cnt-(item+1))/item_cnt);      
		document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
		document.getElementById('nav_items').textContent=  (item + 1) + ' / ' + (item_cnt - item-1);  
	}
}


/* set absolute positions (left, top, width, height) for can_m object	
   canvas => mousemove event	=> a pointing device is moved over an element (fired continuously as the mouse moves).
	 this event is fired by 2 html canvas elements => document.getElementById('can_s') and document.getElementById('can_s') */
function mouseMove(e)
{  
  if ( is_task ) 
	{   	
		let can = e.target || e.srcElement;  			// source canvas element object reference  		
		let dx =  can.width / stim[item].length;  // ex getCellFromXY()
		let col = Math.floor ( e.offsetX / dx );  // stimulus (character) array index
	
		// set absolute positions (left, top, width, height) for m_canvas object			
		let can_m = document.getElementById('can_m');
		can_m.style.left= col * dx  + 'px';  // top and left properties require a unit of measure, not just a number
		can_m.style.top= can.top + 'px';    
		can_m.width= dx;  
		can_m.height= can.height;  
  
		//if ( can_m.style.display != 'block' )  {  // ? performance
		  can_m.style.display= 'block';	
		//}	
		let data = [can.id, col];  //  xxx
		can_m.setAttribute('data-cell', data.join(';'));   // 'data-cell' = 'canvas.id;column index'  
				
		let ctx= can_m.getContext('2d'), width = can_m.width, height= can_m.height, txt;
		let r = Math.floor( Math.min(width,height)/2 );   
		ctx.font = r.toString() + 'px ' + font_family;   	// update font	
		
		ctx.textAlign = 'center';    // ctx.textAlign = "left" || "right" || "center" || "start" || "end";   Default value: 	start
	  ctx.textBaseline = 'middle';  // ctx.textBaseline = "top" || "hanging" || "middle" || "alphabetic" || "ideographic" || "bottom";    Default value: alphabetic
				
		if ( can.id == 'can_s' )  {	
		  txt = stim[item][col];
			ctx.fillStyle= bclr_s;
			ctx.fillRect(0, 0, width, height);    
			ctx.fillStyle= pclr_s;
		}
		else  {
			ctx.fillStyle= bclr_d;
			ctx.fillRect(0, 0, width, height); 
			ctx.fillStyle= pclr_d;
			txt = ans[item][col];
		}	
		ctx.fillText(txt, Math.round(width/2), Math.round(height/2));				
	}	
}


// ondragstart event for html can_m object 
// html => <canvas id="can_m" draggable="true" ondragstart="dragStart(event)"></canvas> 
function dragStart(e)  {    
	// real MIME type "text/plain" being unsupported in IE11
	if ( is_task )  {  // data = [can.id, col]
		e.dataTransfer.setData('text', document.getElementById('can_m').getAttribute('data-cell') );	//  [0] = canvas.id; [1]= cell index
	}
}


// ondragover() event for html canvas element/object
// the ondragover event specifies wher the dragged data can be dropped
// html:  <canvas id="can_s" onmousemove="mouseMove(event)" ondragover="allowDrop(event)" ondrop="drop(event)"></canvas> /
function allowDrop(e)  {
  if ( is_task )  {
		e.preventDefault();
	}
}


// ondrop() event for html canvas element/object
// html: <canvas id="can_s" onmousemove="mouseMove(event)" ondragover="allowDrop(event)" ondrop="drop(event)"></canvas> 
function drop(e)  
{  
  if ( is_task )
	{
		e.preventDefault();   
		
				// data[0] => id of source canvas element; data[1] => source cell column index		
		    // real MIME type "text/plain" being unsupported in IE11
		let data= e.dataTransfer.getData('text').split(';');  //  data from start or SOURCE rect => source cell position 		
		let s_col= data[1];  // source cell index		
		
		let d_can = e.target || e.srcElement;  			// destination canvas element object reference  			
		let d_col = Math.floor ( e.offsetX / ( d_can.width / stim[item].length ) );  // destination character array index
				
		// update (swap) user answer => swap rectangles/cells positions
		if ( data[0] == 'can_s' && d_can.id == 'can_d' ) {  // Down row => Up row drag & drop = stim <=> ans
			let tmp= stim[item][s_col];
			stim[item][s_col]= ans[item][d_col];
			ans[item][d_col]= tmp;
		}
		else if ( data[0] == 'can_d' && d_can.id == 'can_s' ) {  // Up => Down drag & drop = ans <=> stim
			let tmp= ans[item][s_col];
			ans[item][s_col]= stim[item][d_col];
			stim[item][d_col]= tmp;
		}
		else if ( data[0] == 'can_d' && d_can.id == 'can_d' && s_col != d_col ) {  // Up => Up drag & drop = ans <=> ans
			let tmp= ans[item][s_col];
			ans[item][s_col]= ans[item][d_col];
			ans[item][d_col]= tmp;
		}
		//else if ( data[0] == 'can_s' && d_can.id == 'can_s' ) {  // Do nothing for Down row => Down row drag & drop		

		drawStim();  // repaint stimulus
	}
}


// set default or user-loaded (required) letters
function setChar()
{
	let size= Object.keys(letters).length;
	document.getElementById('c_min').max= size;  // .min is always 3
	document.getElementById('c_max').max= size;  // .min is always 3
	if ( parseInt(document.getElementById('c_min').value, 10) > size ) {
		document.getElementById('c_min').value= 3;  
	}
  if ( parseInt(document.getElementById('c_max').value, 10) > size ) {
		document.getElementById('c_max').value= size;
	}
	
	let cont= document.getElementById('char_req');
	cont.innerHTML= '';
	for (let key in letters)  {
		let check = document.createElement('input');
    check.type = 'checkbox';
    check.id = 'creq' + key;
		check.setAttribute('data-value', key);
    check.setAttribute('class', 'char-req');
		check.value = key;
    check.title= key;
 
    let lab = document.createElement('label');
    lab.htmlFor = check.id;
    lab.title= key;
		lab.appendChild(document.createTextNode(key));
    
    cont.appendChild(check);
    cont.appendChild(lab);
		
		//let br = document.createElement('br');
    //cont.appendChild(br);
		cont.innerHTML += '&Tab;';  // or: '&#9;' or: '&#x9;' or: '&Tab;'
		
	}
}


// https://developer.mozilla.org/en-US/docs/Web/API/FileReader/readAsText
// https://stackoverflow.com/questions/23331546/how-to-use-javascript-to-read-local-text-file-and-read-line-by-line
// Loads alphabetic letters from a text file
function loadLetters() {
  let file = document.querySelector('#file_let').files[0];
  let delimit = document.getElementById('c_del_l').checked ? '\n' : document.getElementById('delim_l').value;
	let reader = new FileReader(); 
	reader.addEventListener('load', function () 
	{	   
		let chars = reader.result.split(delimit);  // xxx new array (reader.result == a complete text file) 
		let len = chars.length;
		if ( len < 1 ) {
			alert('The number of alphabet letters loaded is zero. Make sure you are using the correct delimiter for the selected file. Check the uploaded file. Currently selected Delimiter character = ' + delimit);
		}
		else if ( len < 12 || len > 247) {
			alert('The number of letters of the alphabet used must be in this range: 12(Rotokas)..247(Tamil). Check the uploaded file. Currently selected Delimiter character = ' + delimit); 
		}
		else  
		{
			char_def= false;
			letters= {};
			let emp= 0, dup= 0;
			for (let i= 0; i< len; i++)  
			{
				key= chars[i].trim();			
				if ( chars[i] ==  '' ) {
					emp++;
				}
				else if ( letters[key] || letters[key.toLocaleLowerCase()] || letters[key.toLocaleUpperCase()] )  {
					dup++;
				}
				else {
					letters[key] = 1;
				}
			}
			let let_cnt= Object.keys(letters).length;
			document.getElementById('letters_cnt').textContent= let_cnt + ' (custom Alphabet)';  // Number of letters of the alphabet (26 or loaded from txt file) 
			setChar();	
      alert('File upload complete: ' + file.name + ' \n Total number of loaded letters of the alphabet (characters) = ' + let_cnt + ' \n Number of blank (empty) characters skipped = '  + emp + ' \n Number of letters (characters) repeated = ' + dup + ' \n Total number of non-loaded characters = ' + (emp + dup) ); 
			
		}		
  }, false);
  /*reader.addEventListener('error', function() {
    alert('Error occurred reading file: ' + file.name);
  });*/

  if (file) {
    reader.readAsText(file);
  }
}


// https://developer.mozilla.org/en-US/docs/Web/API/FileReader/readAsText
// text html: https://www.howtoexcel.org/split-text-by-delimiter/
// Loads the user's dictionary from a text file
function loadDictionary() {
  let file = document.getElementById('file_dic').files[0];
  let delimit = document.getElementById('c_del_w').checked ? '\n' : document.getElementById('delim_w').value;
	let reader = new FileReader(), stimuli, key, emp= 0, dup= 0;
	reader.addEventListener('load', function () 
	{	
	  stimuli = reader.result.split(delimit);
		let tot = stimuli.length;
		if ( tot < 1 ) {
			alert('The number of words (stimuli) loaded is zero. Make sure you are using the correct delimiter for the selected file. Check the uploaded file. Currently selected Delimiter character = ' + delimit);
		}
		else  
		{
			words= {}; 
			for (let i = 0; i < tot; i++)	
			{			
				key = stimuli[i].toString().replace(/\r\n|\n|\r/g, '').trim();  
				if ( key == '' ) {
					emp++;
				}
				else if ( words[key]  )  { 
					dup++;  // escape duplicates
				}
				else  {
					words[key] = 1;
				}
			}		
			let wor_cnt= Object.keys(words).length;   // Number of words loaded from the dictionary
			document.getElementById('words_cnt').textContent= wor_cnt + ' (custom Dictionary)';  // Update number of words loaded from the dictionary
			dic_def= false;   // default Dictionary is not used => user loaded custom Dictionary loaded from txt file 
			
			alert('File upload complete: ' + file.name + ' \n Total number of successfully loaded words (stimuli) = ' + wor_cnt + ' \n Number of empty (blank) words = '  + emp + ' \n Number of words repeated = ' + dup + ' \n Total number of non-loaded words (stimuli) = ' + (emp + dup) ); 
		}
	}, false);
  /*reader.addEventListener('error', function() {
    alert('Error occurred reading file: ' + file.name);
  });*/

  if (file) {
    reader.readAsText(file);
  }
}


// https://www.freecodecamp.org/news/how-to-reverse-a-string-in-javascript-in-3-different-ways-75e4763c68cb/
function reverseString(str) {
    return str.split("").reverse().join("");
}


// check if created word is valid (only when demo == true)
function checkWord()
{
	let txt = ans[item].join(''), e = document.getElementById('feedback'); 
	if ( txt.length == 0 )  {
		e.textContent= "You didn't create a word. Move the letters from the bottom line to the top line so that they form a valid word.";
	}
	else  {
		let res = getScore(item);
		if ( is_ltr == '2' ) {
			txt= reverseString(txt);
		}
		e.textContent=  'Word check: "' + txt + '". ';
		e.textContent+=  res > 0 ? 'The word is found in the used dictionary. Number of points: ' + res  + '.' : 'The word is invalid. Number of points: 0. Minimum validity threshold = ' + limen + '.';		
	}
}	



// return number of points for item i
function getScore(i) 
{
	let txt= ans[i].join(''), txt_low= txt.toLocaleLowerCase(), txt_up= txt.toLocaleUpperCase(), r_txt, r_txt_low, r_txt_up, ok;
	if ( is_ltr != '1' )  {
		r_txt= reverseString(txt), r_txt_low= r_txt.toLocaleLowerCase(), r_txt_up= r_txt.toLocaleUpperCase();
	}
	switch ( is_ltr )  {	
		case '1':  // parse left-to-right
			ok = txt.length >= limen && ( words[txt] == 1 || words[txt_low] == 1 || words[txt_up] == 1 );
	    break;
		case '2':	// parse right-to-left
			ok = r_txt.length >= limen && ( words[r_txt] == 1 || words[r_txt_low] == 1 || words[r_txt_up] == 1 );
			break;
		case '3':	// parse both directions: left OR right
			ok = txt.length >= limen && ( words[txt] == 1 || words[txt_low] == 1 || words[txt_up] == 1 );
			if ( !ok ) {
				ok = r_txt.length >= limen && ( words[r_txt] == 1 || words[r_txt_low] == 1 || words[r_txt_up] == 1 );
			}
			break;	
	}	
	let res = ok == true ? txt.length : 0;
	return res;
}	


// end of task (research)
function theEnd()  
{    
	is_task = false;

	document.getElementById('task_section').style.display= 'none';	
  document.getElementById('end').style.display= 'block';   
  document.getElementById('can_s').style.display= 'none';  
  document.getElementById('can_d').style.display= 'none';  
	document.getElementById('can_m').style.display= 'none';	
	document.getElementById('result_wrapper').style.display= 'block';
  
	let header= [ 'Item', 'Characters', 'Answer', 'Points' ];
	let size= stim.length,  score = 0, max_score = 0, arr = [], rows = [], len, p, s;		
	for (let i = 0; i < size; i++)  {		// for (let i = 0; i < item_cnt; i++)			
		p = getScore(i);   // number of points in this i-item	
		arr.push(p);
		score+= p;  // sum of points in whole task
		len= stim[i].length;
		max_score+= len;
		s= ans[i].join('');  // user answer for this i-item
	  if ( is_ltr == '2' ) {
			s= reverseString(s);
		}
		rows[i]= [ parseInt(i+1,10), len, s, p ]; 
	}  
	rows.push (  [ 'Sum',  max_score, '', score,   ] );   // totals 	
  createTableArray(document.getElementById('table1'), rows, header); 
  document.getElementById('s_limen').textContent= limen;   

	
	// table2 => Descriptive statistics	
	const D = 3;  // number of decimals
	rows = [];  // reset table rows (for table2)
	let n = arr.length;  
	let m =   n > 0  ?  mean(arr).toFixed(D) : '';      
	let sd =  n > 1  ?  stDev(arr, m).toFixed(D) : '';   	 	
  rows.push( [ m, sd, n ] );  // last row
	
	header= ['Average', 'Standard deviation', 'N']; // header		
	createTableArray(document.getElementById('table2'), rows, header); 

  // update 'score' label => above "Repeat" button
  let percent = (100 * score / max_score).toFixed(D);
	document.getElementById('score').textContent= 'Score= ' + score + '. Total number of characters displayed= ' + max_score + '. Percent= ' + percent + '%';
	
	document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
	
	try 
  {
    if (store) { 
			setStorage(score, max_score, percent, m, sd);   			
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }  

}


// return arithmetic mean (mean, average) of input array (arr)
// no validation => input array must be in valid number format	
function mean(arr)  
{	
	let len= arr.length, sum= 0;
	for (let i=0; i<len; i++) {
		sum+= arr[i];
	}
	return (sum/arr.length);
}	


// return Sample Standard Deviation (with Bessel's correction) => N-1
// no validation => input array must be in valid number format
// arr = array argument;  mean = average for input arr 
function stDev(arr, mean) 
{
	let len= arr.length, sum= 0, res;
	for (let i=0; i<len; i++) {
		sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
	}
	res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
	return res;
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	if (dd < 10) { dd='0'+dd; } 
	if (mm < 10) { mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:
	 score = sum (number) of points in task
   max_score = maximum possible points in task
   percent = 100 * points/ max_points	 
   mean = average (mean) for user points 
   sd = standard deviation for user points	 */   
function setStorage(score, max_score, percent, mean, sd)  
{    
	let oldItems = JSON.parse(localStorage.getItem('lpt-v1.1')) || []; 
  let id= oldItems.length + 1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(),  		   
			'score' : score,
			'max-score' : max_score,
			'percent' : percent,
			'mean'  : mean, 	
			'sd' : sd,			
			'letter-min' : c_min,
      'letter-max' : c_max,			
			'iteration' : iter,
			'items' : item_cnt,
			'letter-type' : is_lower,
			'required' : is_req ? 1 : 0,
			'parse' : is_ltr,
			'threshold' : limen,
			'repeat' : duplic ? 1 : 0,
			'sort' : is_sort ? 1 : 0,			
			'timer' : timer_enabled ? timer : '',	
			'is-back' : is_back ? 1 : 0,
			'shuffle' : is_shuffle ? 1 : 0,
			'demo' : demo ? 1 : 0,
      'default-dic' : dic_def  ? 1 : 0,
      'default-let' : char_def ? 1 : 0 			
  };

  oldItems.push(newItem);

  localStorage.setItem('lpt-v1.1', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems); 
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
          col.push(key);
        }
    }
  } 
  let table = document.createElement('table');
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement('th');      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= '';
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement('table');  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement('th');      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('lpt-v1.1')  ?  JSON.parse(localStorage.getItem('lpt-v1.1')).length : 0;                                                        
  if (cnt === 0)  {
    //alert('You have no data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved results:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 { 
   window.localStorage.removeItem('lpt-v1.1'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('lpt-v1.1');
	if (!data)  {
    //alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// https://stackoverflow.com/questions/60698115/how-to-copy-a-javascript-2d-array-to-clipboard-to-paste-it-in-excel
// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// copied from https://stackoverflow.com/questions/400212/how-do-i-copy-to-the-clipboard-in-javascript
// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('lpt-v1.1'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays Latest research results in new empty web page (tab) 
// parameter: id => DOM/html object (div or table)
function newWinLast(id)
{ 	
	let ind = document.getElementById('last_data_sel').selectedIndex;
	ind++;
	let e = document.getElementById('table' + ind);	
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{ 
  let ind = document.getElementById('last_data_sel').selectedIndex, tbl;
  switch (ind) {
    case 0:
			tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
			break;		
    case 1:
			tbl = document.getElementById('table2').getElementsByTagName('table')[0];  
			break;
	}		
  if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// https://stackoverflow.com/questions/3065342/how-do-i-iterate-through-table-rows-and-cells-in-javascript
// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}
